/*
 * Copyright (C) 2013 Philippe Komma
 * 
 * This file is part of k-Means: A C++ clustering library 
 * 
 * k-Means is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Foobar is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with k-Means. If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __MATH_UTILS_H__
#define __MATH_UTILS_H__

#include "Matrix.h"

template <class T> T random()
{
	return static_cast<T>( rand() ) / RAND_MAX;
}

template <class T> class EigenSystem
{
	protected:

		T m_val;
	
		CVector<T> m_vec;

		bool m_converged;
		
		T m_eps;

	public:

		EigenSystem() :
		    m_val( 0 ),
			m_converged( false ),
			m_eps( 0 )
		{
		}

		EigenSystem( T val, CVector<T> const& vec, bool converged, T eps ) :
			m_val( val ),
			m_vec( vec ),
			m_converged( converged ),
			m_eps( eps )
		{
		}		

		T val() {
			return m_val;
		}

		CVector<T>& vec() {
			return m_vec;
		}

		bool converged() {
			return m_converged;
		}

		T eps() {
			return m_eps;
		}
};

template <class T> EigenSystem<T> dominantEigenVector( CMatrix<T> const& M, int maxIter = 100, T eps = 1e-05 )
{	
	CMatrix<T> C( M.cov() );
	CVector<T> v( CVector<T>::random( M.nCols() ) );
	v.normalize();

	int iter = 0;
	bool done = false;
	bool converged = false;
	T dist = 0;
	while ( !done ) {
		CVector<T> current( C * v );
		current.normalize();
		++iter;
		dist = ( v - current ).length();
		converged = dist < eps;
		if ( converged || ( iter >= maxIter ) ) {
			done = true;			
		}
		v = current;
	}
	if ( converged ) {
		return EigenSystem<T>( ( C * v ) * v, v, true, dist );
	} else {
		return EigenSystem<T>();
	}
}

#endif