/*
 * Copyright (C) 2013 Philippe Komma
 * 
 * This file is part of k-Means: A C++ clustering library 
 * 
 * k-Means is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Foobar is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with k-Means. If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __MATRIX_H__
#define __MATRIX_H__

#include <bitset>

#include "Vector.h"
#include "sseUtils.h"
#include "MemUtils.h"

template <class T> class CIndexValuePair 
{
	protected:

		CVector<int> m_idx;
	
		CVector<T> m_val;

	public:

		CIndexValuePair( int n ) :
			m_idx( n ),
			m_val( n )
		{
		}

		~CIndexValuePair() {
		}

		CVector<int>& idx() {
			return m_idx;
		}

		CVector<T>& val() {
			return m_val;
		}
};

template <class T> class CMatrix
{
	protected:

		int m_nRows;
	
		int m_nCols;

		T* m_x;

	public:	

		CMatrix() :
			m_nRows( 0 ),
			m_nCols( 0 ),
			m_x( NULL ) 
		{
		}

		CMatrix( int nRows, int nCols, bool init = false, T initValue = 0 )	{
			alloc( nRows, nCols, init, initValue );
		}

		CMatrix( CMatrix<T> const& M ) {
			alloc( M.nRows(), M.nCols() );
			memcpy( dataPtr(), M.dataPtr(), M.nEntries() * sizeof( T ) );
		}

		void alloc( int nRows, int nCols, bool init = false, T initValue = 0 ) {
			m_nRows = nRows;
			m_nCols = nCols;			
			m_x = CMemory<T>::alloc( nRows * nCols );
			if ( init ) {
				if ( initValue == 0 ) {
					zeros();
				} else {
					fill( initValue );
				}
			}
		}

		void release() {
			if ( m_x ) {
				CMemory<T>::release( m_x );
			}
			m_nRows = 0;
			m_nCols = 0;
		}

		int nRows() const {
			return m_nRows;
		}

		int nCols() const {
			return m_nCols;
		}

		int nEntries() const {
			return nRows() * nCols();
		}

		T* dataPtr() const {
			return m_x;
		}

		T* operator[]( int row ) const {
			return &m_x[ row * m_nCols ];
		}

		CMatrix<T> operator*( T f ) {
			CMatrix<T> result( nRows(), nCols() );
			int i;
			T* src = dataPtr();
			T* dst = result.dataPtr();
			for ( i = 0; i < nRows() * nCols(); ++i ) {
				*dst = ( *src ) * f;
				++src;
				++dst;
			}
			return result;
		}

		CVector<T> operator*( CVector<T> const& v ) {
			CVector<T> result( nRows() );
			int r;
			T* ptr = dataPtr();
			for ( r = 0; r < nRows(); ++r, ptr += nCols() ) {
				result[ r ] = CVector<T>( nCols(), ptr ) * v;
			}
			return result;
		}

		CMatrix<T> operator*( CMatrix<T> const& m ) {
			CMatrix<T> aux( m );
			aux.transpose();
			CMatrix<T> result( nRows(), m.nCols() );
			int r, c;
			T* srcPtrA = dataPtr();
			T* srcPtrB = aux.dataPtr();
			T* dstPtr = result.dataPtr();
			for ( r = 0; r < nRows(); ++r ) {
				for ( c = 0; c < m.nCols(); ++c ) {
					*dstPtr = CVector<T>( nCols(), srcPtrA ) * CVector<T>( m.nRows(), srcPtrB );
					srcPtrB += m.nRows();
					++dstPtr;
				}
				srcPtrA += nCols();
				srcPtrB = aux.dataPtr();
			}
			return result;
		}

		CVector<T> rowVector( int row ) {
			return CVector<T>( nCols(), m_x + row * nCols() );
		}

		friend std::ostream& operator<<( std::ostream& out, CMatrix<T>& mat ) {
			int r;
			out << "[" << endl;
			for ( r = 0; r < mat.nRows(); ++r ) {
				out << "  " << CVector<T>( mat.nCols(), mat[ r ] ) << endl;
			}
			out << "]" << endl;
			return out;
		}

		void zeros() {
			memset( m_x, 0, nRows() * nCols() * sizeof( T ) );
		}

		void fill( T initValue ) {
			std::fill( dataPtr(), dataPtr() + nEntries(), initValue );
		}

		void removeRow( int r ) {
			std::copy( dataPtr() + ( r + 1 ) * nCols(), dataPtr() + nEntries(), dataPtr() + r * nCols() );
			m_nRows--;
		}

		void min( int dim, CIndexValuePair<T>& result );

		CIndexValuePair<T> min( int dim );

		void max( int dim, CIndexValuePair<T>& result );

		CIndexValuePair<T> max( int dim );

		void dist( CVector<T> const& vec, CVector<T>& result ) const;

		CVector<T> dist( CVector<T> const& vec ) const;

		void transpose();

		void stdNormalize();

		CMatrix<T> cov() const {
			CMatrix<T> aux( *this );

			aux.transpose();

			int r, c;
			T* ptr_x = aux.dataPtr();
			for ( r = 0; r < aux.nRows(); r++ ) {
				const CVector<T> v( aux.nCols(), ptr_x );
				T avg = v.mean();
				for ( c = 0; c < aux.nCols(); c++ ) {
					*ptr_x = *ptr_x - avg;
					++ptr_x;
				}
			}

			CMatrix<T> transposed( aux );

			aux.transpose();

			return transposed * aux * ( static_cast<T>( 1.0 ) / ( nRows() - 1 ) );			
		}
};

template <class T> void CMatrix<T>::min( int dim, CIndexValuePair<T>& result )
{
	if ( dim == 1 ) {
		transpose();
	}
	
	int r;
	T* src = dataPtr();
	T* valPtr = result.val().dataPtr();
	int* idxPtr = result.idx().dataPtr();
	for ( r = 0; r < nRows(); ++r, src += nCols(), ++idxPtr, ++valPtr ) {
		std::pair<int, T> min = CVector<T>( nCols(), src ).min();
		*idxPtr = min.first;
		*valPtr = min.second;		
	}
	
	if ( dim == 1 ) {
		transpose();
	}
}

template <class T> CIndexValuePair<T> CMatrix<T>::min( int dim )
{
	int n;
	if ( dim == 1 ) {
		n = nCols();
	} else {
		n = nRows();
	}

	CIndexValuePair<T> result( n );

	min( dim, result );

	return result;
}

template <class T> void CMatrix<T>::max( int dim, CIndexValuePair<T>& result )
{
	if ( dim == 1 ) {
		transpose();
	}
	
	int r;
	T* src = dataPtr();
	T* valPtr = result.val().dataPtr();
	int* idxPtr = result.idx().dataPtr();
	for ( r = 0; r < nRows(); ++r, src += nCols(), ++idxPtr, ++valPtr ) {
		std::pair<int, T> max = CVector<T>( nCols(), src ).max();
		*idxPtr = max.first;
		*valPtr = max.second;		
	}
	
	if ( dim == 1 ) {
		transpose();
	}
}

template <class T> CIndexValuePair<T> CMatrix<T>::max( int dim )
{
	int n;
	if ( dim == 1 ) {
		n = nCols();
	} else {
		n = nRows();
	}

	CIndexValuePair<T> result( n );

	max( dim, result );

	return result;
}

template <class T> void CMatrix<T>::dist( CVector<T> const& vec, CVector<T>& result ) const
{
	int i;
	T* src = dataPtr();
	T* dst = result.dataPtr();
	for ( i = 0; i < nRows(); ++i, ++dst, src += nCols() ) {
		*dst = CVector<T>( nCols(), src ).dist( vec );
	}
}

template <class T> CVector<T> CMatrix<T>::dist( CVector<T> const& vec ) const
{
	CVector<T> result( nRows() );

	dist( vec, result );

	return result;
}

template <class T> void CMatrix<T>::transpose()
{
	CMatrix<T> aux( nCols(), nRows() );

	int r, c;
	T* src = dataPtr();
	T* dst = aux.dataPtr();
	for ( c = 0; c < aux.nCols(); ++c ) {
		for ( r = 0; r < aux.nRows(); ++r ) {
			dst[ r * aux.nCols() + c ] = *src;
			++src;
		}
	}

	memcpy( dataPtr(), aux.dataPtr(), nCols() * nRows() * sizeof( T ) );

	std::swap( m_nRows, m_nCols );
}

template <class T> void CMatrix<T>::stdNormalize()
{
	transpose();

	int r, c;
	T* ptr_x = m_x;
	for ( r = 0; r < m_nRows; r++ ) {
		const CVector<T> v( m_nCols, ptr_x );
		T avg = v.mean();
		T std = v.std( avg );
		for ( c = 0; c < m_nCols; c++ ) {
			*ptr_x = ( *ptr_x - avg ) / std;
			++ptr_x;
		}
	}

	transpose();
}

#endif