/*
 * Copyright (C) 2013 Philippe Komma
 * 
 * This file is part of k-Means: A C++ clustering library 
 * 
 * k-Means is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Foobar is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with k-Means. If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __VECTOR_H__
#define __VECTOR_H__

#include <cmath>
#include <algorithm>

#include "MemUtils.h"
// #include "MathUtils.h"

#ifdef min
	#undef min
#endif

#ifdef max
	#undef max
#endif

template <class T> class CVector;

template <class T> class CSortResult
{
	protected:

		CVector<T> m_val;
		
		CVector<int> m_idx;

	public:

		CSortResult( int n ) :
		    m_val( n ),
			m_idx( n )
		{
		}
		
		CVector<T>& val() {
			return m_val;
		}

		CVector<int>& idx() {
			return m_idx;
		}
};

template <class T> class CVector
{
	protected:

		int m_n;

		T* m_x;

		bool m_deletable;

	public:

		CVector() :
			m_n( 0 ),
			m_x( NULL ),
			m_deletable( false )
		{
		}

		CVector( int n, T* ptr ) :
		    m_n( n ),
			m_x( ptr ),
			m_deletable( false )
		{
		}

		CVector( int n, bool init = false ) {
			m_x = alloc( n, init );
		}

		CVector( int lo, int hi ) {
			alloc( hi - lo + 1 );
			int i;
			T* ptr = m_x;
			for ( i = lo; i <= hi; ++i ) {
				*ptr = static_cast<T>( i );
				++ptr;
			}
		}

		CVector( CVector<T> const& v ) {
			alloc( v.size() );
			memcpy( dataPtr(), v.dataPtr(), v.size() * sizeof( T ) );
		}

		~CVector()
		{
			release();
		}

		T* alloc( int n, bool init = false ) {
			m_n = n;
			m_deletable = true;
			m_x = CMemory<T>::alloc( n );
			if ( init ) {
				zeros();
			}
			return m_x;
		}

		void release() {
			m_n = 0;
			if ( m_deletable ) {
				CMemory<T>::release( m_x );
			} else {
				m_x = NULL;
			}
			m_deletable = false;			
		}

		void setDataPtr( T* ptr ) {
			m_x = ptr;
		}

		void incDataPtr( int numEntries ) {
			m_x += numEntries;
		}

		T operator[]( int idx ) const {
			return m_x[ idx ];
		}

		T& operator[]( int idx ) {
			return m_x[ idx ];
		}

		int size() const {
			return m_n;
		}

		T* dataPtr() const {
			return m_x;
		}

		bool deletable() const {
			return m_deletable;
		}

		void removeEntry( int idx ) {
			std::copy( dataPtr() + idx + 1, dataPtr() + size(), dataPtr() + idx );
			--m_n;
		}

		T mean() const {
			int i;
			T sum = 0;
			T* ptr_x = m_x;
			for ( i = 0; i < m_n; i++ ) {
				sum += *ptr_x;
				++ptr_x;
			}
			return sum / m_n;
		}

		T std( T const& mean ) const {
			int i;
			T sum = 0;
			T* ptr_x = m_x;
			for ( i = 0; i < m_n; i++ ) {
				const float d = ( *ptr_x - mean );
				sum += d * d;
				++ptr_x;
			}
			sum /= m_n - 1;
			return sqrt( sum );
		}

		T std() const {
			return std( mean() );
		}

		T length() const {
			int i;
			T sum = 0;
			T* ptrX = m_x;
			for ( i = 0; i < m_n; ++i, ++ptrX ) {
				sum += ( *ptrX ) * ( *ptrX );
			}
			return sqrt( sum );
		}

		void normalize() {
			T len = length();
			if ( len > 1e-12 ) {
				int i;
				T* ptrX = m_x;
				for ( i = 0; i < m_n; ++i, ++ptrX ) {
					*ptrX /= len;
				}
			}
		}

		void zeros() {
			memset( m_x, 0, m_n * sizeof( T ) );
		}

		static CVector<T> random( int n ) {
			CVector<T> v( n );

			int i;
			for ( i = 0; i < v.size(); ++i ) {
				v[ i ] = ::random<T>();
			}

			return v;
		}

		float operator*( CVector<T> const& v ) {
			int i;
			T result = 0;
			T* src = dataPtr();
			T* dst = v.dataPtr();
			for ( i = 0; i < m_n; ++i, ++src, ++dst ) {
				result += ( *src ) * ( *dst );
			}
			return result;
		}

		void operator+=( CVector<T> const& v ) {
			int i;
			T* dst = m_x;
			T* src = v.dataPtr();
			for ( i = 0; i < m_n; ++i ) {
				*dst += *src;
				++src;
				++dst;
			}
		}

		void operator*=( const T fac ) {
			int i;
			T* dst = m_x;
			for ( i = 0; i < m_n; ++i, ++dst ) {
				*dst *= fac;
			}
		}

		CVector<T> operator-( CVector<T> const& v ) {
			CVector<T> result( v.size() );
			int i;
			T* srcA = dataPtr();
			T* srcB = v.dataPtr();
			T* dst = result.dataPtr();
			for ( i = 0; i < v.size(); ++i ) {
				*dst = ( *srcA ) - ( *srcB );
				++srcA;
				++srcB;
				++dst;
			}
			return result;
		}

		void operator=( CVector<T> const& v ) {
			memcpy( m_x, v.dataPtr(), size() * sizeof( T ) );
		}

		T dist( CVector<T> const& v ) {
			int i;
			T sum = 0;
			T* ptrA = dataPtr();
			T* ptrB = v.dataPtr();
			for ( i = 0; i < m_n; ++i ) {
				T diff = *ptrA - *ptrB;
				diff *= diff;
				sum += diff;
				++ptrA;
				++ptrB;
			}
			return sqrt( sum );
		}

		std::pair<int, T> min() const {
			T* ptr_x = m_x;
			int min_idx = 0;
			T min_val = *ptr_x;
			
			int i;
			ptr_x++;
			for ( i = 1; i < m_n; ++i, ++ptr_x ) {
				if ( *ptr_x < min_val ) {
					min_idx = i;
					min_val = *ptr_x;
				}
			}

			return std::pair<int, T>( min_idx, min_val );
		}

		std::pair<int, T> max() const {
			T* ptr_x = m_x;
			int max_idx = 0;
			T max_val = *ptr_x;
			
			int i;
			ptr_x++;
			for ( i = 1; i < m_n; ++i, ++ptr_x ) {
				if ( *ptr_x > max_val ) {
					max_idx = i;
					max_val = *ptr_x;
				}
			}

			return std::pair<int, T>( max_idx, max_val );
		}

		CSortResult<T> sort() {
			return quickSort( *this );
		}

		friend std::ostream& operator<<( std::ostream& out, const CVector<T>& v ) {
			int i;
			out << "(";
			for ( i = 0; i < v.size() - 1; ++i ) {
				out << v[ i ] << ",";
			}
			out << v[ v.size() - 1 ] << ")";
			return out;
		}
};

#endif